from cliquet.utils import strip_uri_prefix, Enum, current_service


ACTIONS = Enum(CREATE='create',
               DELETE='delete',
               READ='read',
               UPDATE='update')


class ResourceChanged(object):
    """Triggered when a resource is changed.
    """
    def __init__(self, action, resource, request):
        self.request = request
        service = current_service(request)
        resource_name = service.viewset.get_name(resource.__class__)

        self.payload = {'timestamp': resource.timestamp,
                        'action': action,
                        'uri': strip_uri_prefix(request.path),
                        'user_id': request.prefixed_userid,
                        'resource_name': resource_name}

        matchdict = dict(request.matchdict)

        if 'id' in request.matchdict:
            matchdict[resource_name + '_id'] = matchdict.pop('id')

        self.payload.update(**matchdict)



import contextlib
import warnings

from six.moves.urllib import parse as urlparse

from cliquet import logger
from cliquet.storage import exceptions
from cliquet.utils import json, psycopg2

if psycopg2:
    import psycopg2.extras
    import psycopg2.pool

    psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
    psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY)


class PostgreSQLClient(object):

    pool = None

    def __init__(self, *args, **kwargs):
        if psycopg2 is None:
            message = "You must install psycopg2 to use the postgresql backend"
            raise ImportWarning(message)

        pool_size = kwargs.pop('pool_size')
        self._conn_kwargs = kwargs
        pool_klass = psycopg2.pool.ThreadedConnectionPool
        if PostgreSQLClient.pool is None:
            PostgreSQLClient.pool = pool_klass(minconn=pool_size,
                                               maxconn=pool_size,
                                               **self._conn_kwargs)
        elif pool_size != self.pool.minconn:
            msg = ("Pool size %s ignored for PostgreSQL backend "
                   "(Already set to %s).") % (pool_size, self.pool.minconn)
            warnings.warn(msg)

        # When fsync setting is off, like on TravisCI or in during development,
        # cliquet some storage tests fail because commits are not applied
        # accross every opened connections.
        # XXX: find a proper solution to support fsync off.
        # Meanhwile, disable connection pooling to prevent test suite failures.
        self._always_close = False
        with self.connect(readonly=True) as cursor:
            cursor.execute("SELECT current_setting('fsync');")
            fsync = cursor.fetchone()[0]
            if fsync == 'off':  # pragma: no cover
                warnings.warn('Option fsync = off detected. Disable pooling.')
                self._always_close = True
        # Register ujson, globally for all futur cursors
        with self.connect() as cursor:
            psycopg2.extras.register_json(cursor,
                                          globally=True,
                                          loads=json.loads)

    @contextlib.contextmanager
    def connect(self, readonly=False):
        """Connect to the database and instantiates a cursor.
        At exiting the context manager, a COMMIT is performed on the current
        transaction if everything went well. Otherwise transaction is ROLLBACK,
        and everything cleaned up.

        If the database could not be be reached a 503 error is raised.
        """
        conn = None
        cursor = None
        try:
            conn = self.pool.getconn()
            conn.autocommit = readonly
            options = dict(cursor_factory=psycopg2.extras.DictCursor)
            cursor = conn.cursor(**options)
            # Start context
            yield cursor
            # End context
            if not readonly:
                conn.commit()
        except psycopg2.Error as e:
            if cursor:
                logger.debug(cursor.query)
            logger.error(e)
            if conn and not conn.closed:
                conn.rollback()
            raise exceptions.BackendError(original=e)
        finally:
            if cursor:
                cursor.close()
            if conn and not conn.closed:
                self.pool.putconn(conn, close=self._always_close)


def create_from_config(config, prefix=''):
    """Create a PostgreSQLClient client using settings in the provided config.
    """
    settings = config.get_settings()
    pool_size = int(settings[prefix + 'pool_size'])
    uri = settings[prefix + 'url']
    parsed = urlparse.urlparse(uri)
    conn_kwargs = dict(host=parsed.hostname,
                       port=parsed.port,
                       user=parsed.username,
                       password=parsed.password,
                       database=parsed.path[1:] if parsed.path else '')
    # Filter specified values only, to preserve PostgreSQL defaults
    conn_kwargs = dict([(k, v) for k, v in conn_kwargs.items() if v])
    return PostgreSQLClient(pool_size=pool_size, **conn_kwargs)



# -*- coding: utf-8 -*-
import mock
from pyramid import testing
import uuid
import redis
import json
from datetime import datetime
from contextlib import contextmanager

from cliquet import initialization
from cliquet.events import ResourceChanged, ACTIONS
from cliquet.listeners import ListenerBase
from .support import unittest


class ListenerSetupTest(unittest.TestCase):
    def test_redis_listener_is_enabled_via_setting(self):
        listener = 'cliquet.listeners.redis'
        redis_class = mock.patch(listener + '.RedisListener')
        config = testing.setUp()
        with mock.patch.dict(config.registry.settings,
                             [('event_listeners', listener)]):
            with redis_class as redis_mocked:
                initialization.setup_listeners(config)
                self.assertTrue(redis_mocked.called)


@contextmanager
def broken_redis():
    from redis import StrictRedis
    old = StrictRedis.lpush

    def push(*args, **kwargs):
        raise Exception('boom')

    StrictRedis.lpush = push
    yield
    StrictRedis.lpush = old

UID = str(uuid.uuid4())


class Resource(object):
    record_id = UID
    timestamp = 123456789


class ViewSet(object):
    def get_name(*args, **kw):
        return 'collection'


class Service(object):
    viewset = ViewSet()


class Match(object):
    cornice_services = {'watev': Service()}
    pattern = 'watev'


class Request(object):
    path = '/1/bucket/collection/'
    prefixed_userid = 'tarek'
    matchdict = {'id': UID}
    registry = matched_route = Match()


class ListenerCalledTest(unittest.TestCase):

    def setUp(self):
        self.config = testing.setUp()
        pool = redis.BlockingConnectionPool(max_connections=1,
                                            host='localhost', port=6379, db=0)
        self._redis = redis.StrictRedis(connection_pool=pool)
        self._size = 0

    def _save_redis(self):
        self._size = self._redis.llen('cliquet.events')

    def has_redis_changed(self):
        return self._redis.llen('cliquet.events') > self._size

    def notify(self, event):
        self._save_redis()
        self.config.registry.notify(event)

    @contextmanager
    def redis_listening(self):
        config = self.config
        listener = 'cliquet.listeners.redis'

        # setting up the redis listener
        with mock.patch.dict(config.registry.settings,
                             [('event_listeners', listener),
                              ('event_listeners.redis.pool_size', '1')]):
            initialization.setup_listeners(config)
            config.commit()
            yield

    def test_redis_is_notified(self):
        with self.redis_listening():
            # let's trigger an event
            event = ResourceChanged('create', Resource(), Request())
            self.notify(event)
            self.assertTrue(self.has_redis_changed())

        # okay, we should have the first event in Redis
        last = self._redis.lpop('cliquet.events')
        last = json.loads(last.decode('utf8'))
        self.assertEqual(last['action'], ACTIONS.CREATE)

    def test_notification_is_broken(self):
        with self.redis_listening():
            # an event with a bad JSON should silently break and send nothing
            res = Resource()
            # date time objects cannot be dumped
            res.timestamp = datetime.now()
            event2 = ResourceChanged('create', res, Request())
            self.notify(event2)
            self.assertFalse(self.has_redis_changed())

    def test_redis_is_broken(self):
        with self.redis_listening():
            # if the redis call fails, same deal: we should ignore it
            self._save_redis()

            with broken_redis():
                event = ResourceChanged('create', Resource(), Request())
                self.config.registry.notify(event)

            self.assertFalse(self.has_redis_changed())


class ListenerBaseTest(unittest.TestCase):

    def test_not_implemented(self):
        # make sure we can't use the base listener
        listener = ListenerBase()
        self.assertRaises(NotImplementedError, listener, object())



from __future__ import absolute_import

import redis
from six.moves.urllib import parse as urlparse

from cliquet.cache import CacheBase
from cliquet.storage.redis import wrap_redis_error
from cliquet.utils import json


class Cache(CacheBase):
    """Cache backend implementation using Redis.

    Enable in configuration::

        cliquet.cache_backend = cliquet.cache.redis

    *(Optional)* Instance location URI can be customized::

        cliquet.cache_url = redis://localhost:6379/1

    A threaded connection pool is enabled by default::

        cliquet.cache_pool_size = 50

    :noindex:
    """

    def __init__(self, *args, **kwargs):
        super(Cache, self).__init__(*args, **kwargs)
        connection_pool = redis.BlockingConnectionPool(**kwargs)
        self._client = redis.StrictRedis(connection_pool=connection_pool)

    def initialize_schema(self):
        # Nothing to do.
        pass

    @wrap_redis_error
    def flush(self):
        self._client.flushdb()

    @wrap_redis_error
    def ttl(self, key):
        return self._client.ttl(key)

    @wrap_redis_error
    def expire(self, key, ttl):
        self._client.pexpire(key, int(ttl * 1000))

    @wrap_redis_error
    def set(self, key, value, ttl=None):
        value = json.dumps(value)
        if ttl:
            self._client.psetex(key, int(ttl * 1000), value)
        else:
            self._client.set(key, value)

    @wrap_redis_error
    def get(self, key):
        value = self._client.get(key)
        if value:
            value = value.decode('utf-8')
            return json.loads(value)

    @wrap_redis_error
    def delete(self, key):
        self._client.delete(key)


def load_from_config(config):
    settings = config.get_settings()
    uri = settings['cache_url']
    uri = urlparse.urlparse(uri)
    pool_size = int(settings['cache_pool_size'])

    return Cache(max_connections=pool_size,
                 host=uri.hostname or 'localhost',
                 port=uri.port or 6379,
                 password=uri.password or None,
                 db=int(uri.path[1:]) if uri.path else 0)



from __future__ import absolute_import
from six.moves.urllib import parse as urlparse
import redis
import json

from cliquet.listeners import ListenerBase
from cliquet import logger


class RedisListener(ListenerBase):
    def __init__(self, *args, **kwargs):
        super(RedisListener, self).__init__(*args, **kwargs)
        self.listname = kwargs.pop('listname')
        connection_pool = redis.BlockingConnectionPool(**kwargs)
        self._client = redis.StrictRedis(connection_pool=connection_pool)

    def __call__(self, event):
        try:
            payload = json.dumps(event.payload)
        except TypeError:
            logger.error("Unable to dump the payload", exc_info=True)
            return

        try:
            self._client.lpush(self.listname, payload)
        except Exception:
            logger.error("Unable to send the payload to Redis", exc_info=True)


def _get_options(config, prefix='event_listeners.redis.'):
    options = {}

    for name, value in config.get_settings().items():
        if not name.startswith(prefix):
            continue
        options[name[len(prefix):]] = value

    return options


def load_from_config(config):
    options = _get_options(config)

    uri = options.get('url', 'http://localhost:6379')
    uri = urlparse.urlparse(uri)
    pool_size = int(options.get('pool_size', 1))
    listname = options.get('listname', 'cliquet.events')

    return RedisListener(listname=listname,
                         max_connections=pool_size,
                         host=uri.hostname or 'localhost',
                         port=uri.port or 6379,
                         password=uri.password or None,
                         db=int(uri.path[1:]) if uri.path else 0)
